from locust import HttpUser, between, task, tag


class WebsiteUser(HttpUser):
    """
    Locust stress-testing configuration.
    """

    wait_time = between(2, 4)

    @tag("health")
    @task
    def attempt(self):
        self.client.get("/health")

    @tag("evaluate")
    @task
    def evaluate(self):
        self.client.post(
            "/evaluate/",
            json={
                "id": "test-conversion",
                "control_variant": "a",
                "variants": ["a", "b", "c"],
                "unit_type": "test_unit_type",
                "metrics": [
                    {
                        "id": 1,
                        "name": "Click-through Rate",
                        "nominator": "count(test_unit_type.unit.click)",
                        "denominator": "count(test_unit_type.global.exposure)",
                    },
                ],
                "checks": [
                    {
                        "id": 1,
                        "name": "SRM",
                        "denominator": "count(test_unit_type.global.exposure)",
                    }
                ],
            },
        )



import pandas as pd

from ..dao import Dao
from ..dao import DaoFactory
from ..experiment import Experiment
from ..experiment import Evaluation
from ..experiment import FilterScope
from .test_data import TestData


class TestDao(Dao):
    def __init__(self, test_data: TestData):
        self.metrics = pd.DataFrame({}, columns=Evaluation.metric_columns())
        self.checks = pd.DataFrame({}, columns=Evaluation.check_columns())
        self.test_data = test_data
        self.goals_agg = self.test_data.load_goals_agg()
        self.goals_unit = self.test_data.load_goals_by_unit()

    def get_agg_goals(self, experiment: Experiment) -> pd.DataFrame:
        goals = self.goals_agg[self.goals_agg.exp_id == experiment.id]

        # We can call experiment in 3 ways:
        # 1. provide no `date_from` and `date_to` to evaluate all the experiment data regardless of any date range
        # 2. provide either `date_from` or `date_to` to limit date range from the bottom or from the top
        # 3. provide both `date_from` or `date_to` signaling request for sequential analysis which needs also `date_for`
        # We take min from date_to and date_for in case we evaluate historical experiment where date_for >> date_to
        # already.
        if experiment.date_from is not None and experiment.date_to is not None:
            goals = goals[
                (goals.date >= experiment.date_from.strftime("%Y-%m-%d"))
                & (goals.date <= experiment.date_to.strftime("%Y-%m-%d"))
                & (goals.date <= experiment.date_for.strftime("%Y-%m-%d"))
            ]
        if experiment.date_from is not None:
            goals = goals[goals.date >= experiment.date_from.strftime("%Y-%m-%d")]
        if experiment.date_to is not None:
            goals = goals[goals.date <= experiment.date_to.strftime("%Y-%m-%d")]

        for f in experiment.filters:
            if f.scope == FilterScope.exposure:
                goals = goals[(goals.goal != "exposure") | (goals[f.dimension].isin(f.value))]
            if f.scope == FilterScope.goal:
                goals = goals[(goals.goal == "exposure") | (goals[f.dimension].isin(f.value))]

        return goals

    def get_unit_goals(self, experiment: Experiment) -> pd.DataFrame:
        return self.goals_unit[self.goals_unit.exp_id == experiment.id]

    def load_evaluations_metrics(self, experiment_id: str) -> pd.DataFrame:
        return self.test_data.load_evaluations_metrics(experiment_id)

    def load_evaluations_checks(self, experiment_id: str) -> pd.DataFrame:
        return self.test_data.load_evaluations_checks(experiment_id)

    def load_evaluations_exposures(self, experiment_id: str) -> pd.DataFrame:
        return self.test_data.load_evaluations_exposures(experiment_id)


class TestDaoFactory(DaoFactory):
    def __init__(self, test_data: TestData) -> None:
        self._client = TestDao(test_data)

    def get_dao(self) -> Dao:
        return self._client



from typing import Set
from pyparsing import (
    Word,
    alphas,
    oneOf,
    infixNotation,
    opAssoc,
    nums,
    ParseException,
    alphanums,
)
import pandas as pd


class Parser:
    """
    Parse and evaluate nominator and denominator expressions from goals and give various
    access to parsed forms to e.g. create simple compiler to sql code.
    """

    def __init__(self, nominator: str, denominator: str):
        func = Word(alphas)
        unit_type = Word(alphas + "_").setParseAction(UnitType)
        agg_type = Word(alphas).setParseAction(AggType)
        goal = Word(alphas + "_" + nums).setParseAction(Goal)
        number = Word(nums).setParseAction(Number)
        dimension = Word(alphas + "_").setParseAction(Dimension)
        dimension_value = Word(alphanums + "_" + "-" + "." + "%").setParseAction(DimensionValue)

        ep_goal = (func + "(" + unit_type + "." + agg_type + "." + goal + ")").setParseAction(EpGoal)
        ep_goal_with_dimension = (
            func + "(" + unit_type + "." + agg_type + "." + goal + "(" + dimension + "=" + dimension_value + ")" + ")"
        ).setParseAction(EpGoal)
        operand = number | ep_goal | ep_goal_with_dimension

        multop = oneOf("*")
        divop = oneOf("/")
        plusop = oneOf("+")
        subop = oneOf("-")

        expr = infixNotation(
            operand,
            [
                (multop, 2, opAssoc.LEFT, MultBinOp),
                (divop, 2, opAssoc.LEFT, DivBinOp),
                (plusop, 2, opAssoc.LEFT, PlusBinOp),
                (subop, 2, opAssoc.LEFT, SubBinOp),
            ],
        )

        self._nominator_expr = expr.parseString(nominator)[0]
        self._denominator_expr = expr.parseString(denominator)[0]

    def evaluate_agg(self, goals: pd.DataFrame):
        """
        Get `count`, `sum_value`, `sum_value_sqr` numpy array of shape (variants, metrics) after
        evaluating nominator and denominator expressions from pre-aggregated goals.

        Arguments:
            goals: one row per experiment variant

        See [`Experiment.evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_agg] for details
        on `goals` at input.

        Returns:
            numpy array of shape (variants, metrics) where metrics are in order of
            (count, sum_value, sum_sqr_value)
        """
        sum_value = self._nominator_expr.evaluate_agg(goals)
        sum_sqr_value = self._nominator_expr.evaluate_sqr(goals)
        count = self._denominator_expr.evaluate_agg(goals)
        return count, sum_value, sum_sqr_value

    def evaluate_by_unit(self, goals: pd.DataFrame):
        """
        Get `count`, `sum_value`, `sum_value_sqr` numpy array of shape (variants, metrics) after
        evaluating nominator and denominator expressions from goals aggregated by unit.

        Arguments:
            goals: one row per experiment variant

        See [`Experiment.evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_by_unit] for details
        on `goals` at input.

        Returns:
            numpy array of shape (variants, metrics) where metrics are in order of
            (count, sum_value, sum_sqr_value)
        """
        value_variants, value = self._nominator_expr.evaluate_by_unit(goals, "sum_value")
        value_df = (
            pd.DataFrame(
                {
                    "exp_variant_id": value_variants,  # go through variants
                    "sum_value": value,
                    "sum_sqr_value": value * value,
                }
            )
            .groupby("exp_variant_id")
            .sum()
        )

        count_variants, count = self._denominator_expr.evaluate_by_unit(goals, "count")
        count_df = pd.DataFrame({"exp_variant_id": count_variants, "count": count}).groupby("exp_variant_id").sum()

        df = value_df.join(count_df)
        return df["count"], df["sum_value"], df["sum_sqr_value"]

    def get_goals(self) -> Set:
        """
        Get set of goals that appear in `nominator` and `denominator` expressions as `EpGoal` instances.
        """
        return self._nominator_expr.get_goals().union(self._denominator_expr.get_goals())

    def get_goals_str(self) -> Set[str]:
        """
        Gets set of goals that appear in `nominator` and `denominator` expressions as strings.
        """
        return self._nominator_expr.get_goals_str().union(self._denominator_expr.get_goals_str())


class UnitType:
    def __init__(self, t):
        self.unit_type = t[0]

    def __str__(self):
        return self.unit_type

    __repr__ = __str__


class AggType:
    def __init__(self, t):
        if t[0] not in ["unit", "global"]:
            raise ParseException(f"Only `unit` and `global` aggregation types are supported but `{t[0]}` received.")
        self.agg_type = t[0]

    def __str__(self):
        return self.agg_type

    __repr__ = __str__


class Goal:
    def __init__(self, t):
        self.goal = t[0]

    def __str__(self):
        return self.goal

    __repr__ = __str__


class Dimension:
    def __init__(self, t):
        self.dimension = t[0]

    def __str__(self):
        return self.dimension

    __repr__ = __str__


class DimensionValue:
    def __init__(self, t):
        self.dimension_value = t[0]

    def __str__(self):
        return self.dimension_value

    __repr__ = __str__


class Number:
    def __init__(self, t):
        self.value = float(t[0])

    def __str__(self):
        return f"{self.value}"

    def evaluate_agg(self, goals):
        return self.value

    def evaluate_by_unit(self, goals):
        return self.evaluate_agg(goals)

    def evaluate_sqr(self, goals):
        return self.value * self.value

    def get_goals_str(self) -> Set[str]:
        return set()

    def get_goals(self) -> Set:
        return set()

    __repr__ = __str__


class EpGoal:
    """
    Represents one specification of a goal in nominator or denominator expression
    eg. `count(test_unit_type.global.conversion(product=product_1))` where
    dimensional part is optional.
    """

    def __init__(self, t):
        if t[0] not in ["value", "count", "unique"]:
            raise ParseException(f"Only `value`, `count`, `unique` functions are supported but `{t[0]}` received.")
        if t[0] == "value":
            self.column = "sum_value" if t[0] == "value" else "count"
            self.column_sqr = "sum_sqr_value" if t[0] == "value" else "sum_sqr_count"
        if t[0] == "count":
            self.column = "count"
            self.column_sqr = "sum_sqr_count"
        if t[0] == "unique":
            self.column = "count_unique"
            self.column_sqr = "count_unique"

        self.unit_type = t[2].unit_type
        self.agg_type = t[4].agg_type
        self.goal = t[6].goal
        self.dimension = t[8].dimension if len(t) > 8 else ""
        self.dimension_value = t[10].dimension_value if len(t) > 8 else ""

    def _to_string(self):
        dimension = f"[{self.dimension}={self.dimension_value}]" if self.dimension != "" else ""
        return f"{self.unit_type}.{self.agg_type}.{self.goal}{dimension}"

    def __str__(self):
        return self._to_string()

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self._to_string() == other._to_string()

    def __hash__(self):
        return hash(self._to_string())

    def evaluate_agg(self, goals):
        return self._evaluate_agg(goals, self.column)

    def evaluate_by_unit(self, goals, key):
        g = goals[
            (goals["unit_type"] == self.unit_type)
            & (goals["agg_type"] == self.agg_type)
            & (goals["dimension"] == self.dimension)
            & (goals["dimension_value"] == self.dimension_value)
        ]
        return g["exp_variant_id"], g.xs(key=key, level=1, axis=1)[self.goal]

    def evaluate_sqr(self, goals):
        return self._evaluate_agg(goals, self.column_sqr)

    def _evaluate_agg(self, goals, column):
        return goals[
            (goals["unit_type"] == self.unit_type)
            & (goals["agg_type"] == self.agg_type)
            & (goals["goal"] == self.goal)
            & (goals["dimension"] == self.dimension)
            & (goals["dimension_value"] == self.dimension_value)
        ][column].values

    def get_goals_str(self) -> Set[str]:
        return {self._to_string()}

    def get_goals(self) -> Set:
        return {self}

    def is_dimensional(self):
        return self.dimension != "" and self.dimension_value != ""

    __repr__ = __str__


class BinOp:
    """
    Operation connecting `EpGoal` or `Number` terms in nominator or denominator expression
    eg. `value(test_unit_type.unit.conversion) - value(test_unit_type.unit.refund)`.
    """

    def __init__(self, t):
        self.args = t[0][0::2]

    def symbol(self):
        raise NotImplementedError()

    def evaluate_agg(self, goals):
        raise NotImplementedError()

    def evaluate_by_unit(self, goals):
        raise NotImplementedError()

    def evaluate_sqr(self, goals):
        raise NotImplementedError()

    def get_goals_str(self) -> Set[str]:
        return set().union(*map(lambda o: o.get_goals_str(), self.args))

    def get_goals(self) -> Set[str]:
        return set().union(*map(lambda o: o.get_goals(), self.args))

    def __str__(self):
        sep = f" {self.symbol()} "
        return "(" + sep.join(map(str, self.args)) + ")"

    __repr__ = __str__


class PlusBinOp(BinOp):
    def symbol(self):
        return "+"

    def evaluate_agg(self, goals):
        return self.args[0].evaluate_agg(goals) + self.args[1].evaluate_agg(goals)

    def evaluate_sqr(self, goals):
        return self.args[0].evaluate_sqr(goals) + self.args[1].evaluate_sqr(goals)

    def evaluate_by_unit(self, goals):
        return self.args[0].evaluate_by_unit(goals) + self.args[1].evaluate_by_unit(goals)


class MultBinOp(BinOp):
    def symbol(self):
        return "*"

    def evaluate_agg(self, goals):
        return self.args[0].evaluate_agg(goals) * self.args[1].evaluate_agg(goals)

    def evaluate_sqr(self, goals):
        return self.args[0].evaluate_sqr(goals) * self.args[1].evaluate_sqr(goals)

    def evaluate_by_unit(self, goals):
        return self.args[0].evaluate_by_unit(goals) * self.args[1].evaluate_by_unit(goals)


class DivBinOp(BinOp):
    def symbol(self):
        return "/"

    def evaluate_agg(self, goals):
        return self.args[0].evaluate_agg(goals) / self.args[1].evaluate_agg(goals)

    def evaluate_sqr(self, goals):
        return self.args[0].evaluate_sqr(goals) / self.args[1].evaluate_sqr(goals)

    def evaluate_by_unit(self, goals):
        return self.args[0].evaluate_by_unit(goals) / self.args[1].evaluate_by_unit(goals)


class SubBinOp(BinOp):
    def symbol(self):
        return "-"

    def evaluate_agg(self, goals):
        return self.args[0].evaluate_agg(goals) - self.args[1].evaluate_agg(goals)

    def evaluate_sqr(self, goals):
        return self.args[0].evaluate_sqr(goals) - self.args[1].evaluate_sqr(goals)

    def evaluate_by_unit(self, goals):
        return self.args[0].evaluate_by_unit(goals) - self.args[1].evaluate_by_unit(goals)



import typing
import json as json
from fastapi.responses import JSONResponse


class DataScienceJsonResponse(JSONResponse):
    media_type = "application/json"

    def render(self, content: typing.Any) -> bytes:
        # we need to use default json implementation and explicitly encode in utf-8 to
        # preserver `np.nan` and `np.inf` values.
        return json.dumps(content).encode("utf-8")



import pytz
import pandas as pd


def get_utc_timestamp(dt):
    mytz = pytz.timezone("UTC")
    return mytz.normalize(mytz.localize(dt, is_dst=False))


def goals_wide_to_long(df: pd.DataFrame) -> pd.DataFrame:
    """
    Modify the input DataFrame in a way that it can be evaluatetd using Experiment.evaluate_agg().

    Arguments:
        df: dataframe in wide format - one row per variant and aggregated data in columns

    Returns:
        dataframe in long format - one row per variant and goal

    Input dataframe example:
    ```
    experiment_id   variant_id  views   clicks  conversions     bookings    bookings_squared
    my-exp          a           473661  48194   413             17152       803105
    my-exp          b           471485  47184   360             14503       677178
    my-exp          c           477159  48841   406             15892       711661
    my-exp          d           474934  49090   289             11995       566700
    ```
    """

    # Rename first two columns
    df.columns = ["exp_id", "exp_variant_id"] + df.columns.to_list()[2:]

    # DataFrame `sum_value` to long format
    # Select non squared columns and switch from long to wide
    cols = [col for col in df.columns.to_list()[2:] if "squared" not in col]
    df_long = pd.melt(
        df, id_vars=["exp_id", "exp_variant_id"], value_vars=cols, var_name="goal", value_name="sum_value"
    )

    # DataFrame `sum_sqr_value` to long format
    # Select squared columns and swich from long to wide
    cols_squared = [col for col in df.columns.to_list()[2:] if "squared" in col]
    df_long_sqr = pd.melt(
        df, id_vars=["exp_id", "exp_variant_id"], value_vars=cols_squared, var_name="goal", value_name="sum_sqr_value"
    )
    df_long_sqr["goal"] = df_long_sqr["goal"].apply(_change_variable_name)

    # Merge together and add other necessary columns for evaluation
    goals = pd.merge(left=df_long, right=df_long_sqr, how="outer", on=["exp_id", "exp_variant_id", "goal"])
    goals.insert(2, "unit_type", "test_unit_type")
    goals.insert(3, "agg_type", "global")
    goals.insert(5, "dimension", "")
    goals.insert(6, "dimension_value", "")
    goals.insert(7, "count", 0)
    goals.insert(8, "sum_sqr_count", 0)
    goals.insert(11, "count_unique", 0)
    goals["sum_sqr_value"] = goals.apply(_add_value_squared_where_missing, axis="columns")

    return goals


def _change_variable_name(variable):
    """Change variable name, e.g. `bookings_squared` to `bookings`."""
    return variable.split("_")[0]


def _add_value_squared_where_missing(row):
    """Add values `value_squared` where missing."""
    value_squared = row[-2]
    value = row[-3]

    if value_squared != value_squared:
        return value
    else:
        return value_squared



import pandas as pd
from numpy import allclose
from numpy.testing import assert_array_equal, assert_array_almost_equal

from ..experiment import Experiment, Evaluation
from .test_dao import TestDao
from .test_data import TestData


def evaluate_experiment_agg(experiment: Experiment, test_dao: TestDao):
    goals = test_dao.get_agg_goals(experiment)
    goals = goals[goals.exp_id == experiment.id]

    target = experiment.evaluate_agg(goals)

    assert_experiment(experiment, target, test_dao)

    return target


def evaluate_experiment_simple_agg(experiment: Experiment, test_dao: TestDao):
    """
    This method is used exclusively for test evaluation of an experiment
    with aggregated data in a wide and simple data frame.
    """
    goals = TestData.load_goals_simple_agg()
    target = experiment.evaluate_wide_agg(goals)

    assert_experiment(experiment, target, test_dao)

    return target


def evaluate_experiment_by_unit(experiment: Experiment, test_dao: TestDao):
    goals = test_dao.get_unit_goals(experiment)
    goals = goals[goals.exp_id == experiment.id]

    target = experiment.evaluate_by_unit(goals)

    assert_experiment(experiment, target, test_dao)

    return target


def assert_experiment(experiment: Experiment, target: Evaluation, test_dao: TestDao) -> None:
    metrics = target.metrics
    for m in experiment.metrics:
        assert_metrics(experiment.id, m.id, metrics, test_dao)

    checks = target.checks
    for c in experiment.checks:
        assert_checks(experiment.id, c.id, checks, test_dao)

    assert_exposures(experiment.id, target.exposures, test_dao, unit_type=experiment.unit_type)


def assert_metrics(
    experiment_id: str,
    metric_id: int,
    target: pd.DataFrame,
    test_dao: TestDao,
    precision=3,
) -> None:
    target = target[(target.exp_id == experiment_id) & (target.metric_id == metric_id)]

    expected = test_dao.load_evaluations_metrics(experiment_id)
    expected = expected[expected.metric_id == metric_id]

    assert_array_equal(target.exp_variant_id, expected.exp_variant_id)
    t = target[
        [
            "sum_value",
            "diff",
            "mean",
            "p_value",
            "confidence_interval",
            "confidence_level",
        ]
    ].astype(float)
    atol = 10 ** -precision
    assert allclose(t["sum_value"], expected["sum_value"], atol=atol, equal_nan=True)
    assert allclose(t["diff"], expected["diff"], atol=atol, equal_nan=True)
    assert allclose(t["mean"], expected["mean"], atol=atol, equal_nan=True)
    assert allclose(t["p_value"], expected["p_value"], atol=atol * 10, equal_nan=True)
    assert allclose(t["confidence_interval"], expected["confidence_interval"], atol=atol * 10, equal_nan=True)
    assert allclose(t["confidence_level"], expected["confidence_level"], atol=atol, equal_nan=True)


def assert_checks(
    experiment_id: str,
    check_id: int,
    target: pd.DataFrame,
    test_dao: TestDao,
    precision: int = 4,
) -> None:
    target = target[(target.exp_id == experiment_id) & (target.check_id == check_id)]

    expected = test_dao.load_evaluations_checks(experiment_id)
    expected = expected[expected.check_id == check_id]

    assert_array_equal(target.check_id, expected.check_id)
    assert_array_almost_equal(
        target[target["variable_id"] == "p_value"]["value"],
        expected[expected["variable_id"] == "p_value"]["value"],
        precision,
    )
    assert_array_almost_equal(
        target[target["variable_id"] == "test_stat"]["value"],
        expected[expected["variable_id"] == "test_stat"]["value"],
        precision,
    )
    assert_array_almost_equal(
        target[target["variable_id"] == "confidence_level"]["value"],
        expected[expected["variable_id"] == "confidence_level"]["value"],
        precision,
    )


def assert_exposures(
    experiment_id: str,
    target: pd.DataFrame,
    test_dao: TestDao,
    unit_type: str = "test_unit_type",
    agg_type: str = "global",
) -> None:
    expected = test_dao.load_evaluations_exposures(experiment_id)
    expected = (
        expected[(expected.unit_type == unit_type) & (expected.agg_type == agg_type) & (expected.goal == "exposure")]
        .groupby("exp_variant_id")
        .agg(exposures=("count", "sum"))
        .reset_index()
    )

    assert_array_equal(target.exp_variant_id, expected.exp_variant_id)
    assert_array_equal(target.exposures, expected.exposures)


def check_docstring(doc, indent):
    """
    This function will read through the docstring and grab
    the first python code block. It will try to execute it.
    If it fails, the calling test should raise a flag.
    """
    if not doc:
        return
    start = doc.find("```python\n")
    end = doc.find("```\n")
    if start != -1:
        if end != -1:
            code_part = doc[(start + 10) : end].replace(" " * indent, "")  # noqa: E203
            print(code_part)
            exec(code_part)



from .test_dao import TestDao, TestDaoFactory
from .utils import evaluate_experiment_agg, evaluate_experiment_simple_agg, evaluate_experiment_by_unit
from .utils import assert_experiment, assert_metrics, assert_checks, assert_exposures
from .utils import check_docstring
from .test_data import TestData



import pandas as pd
from typing import List
from pydantic import BaseModel, Field

from ..toolkit import Evaluation
from .req import Experiment, Metric, Check


class MetricStat(BaseModel):
    """
    Per-variant metric evaluation result.
    """

    exp_variant_id: str = Field(title="Variant in the Experiment")
    diff: float = Field(
        title="Difference",
        description="""Relative difference of means of this variant and control variant.
        If this is a variant `b` and `a` is the control variant, then `diff = (b.mean - a.mean) / a.mean`.""",
    )
    mean: float = Field(
        title="Metric Mean",
        description="""Nominator and denominator to calculate the mean
        are given in metric definition. `mean = nominator / denominator`.""",
    )
    sum_value: float = Field(
        title="Metric Value",
        description="""Value of the metric, it is given by the
        nominator in the metric definition.""",
    )
    p_value: float = Field(
        title="p-Value",
        description="""We calculate p-value (under `confidence_level` statistical significance) of the relative
        difference (`diff`) of this variant mean and the control variant mean. We use
        [2-tailed Welch's test](https://en.wikipedia.org/wiki/Welch%27s_t-test)
        with unknown and unequal variance assumption and Welch–Satterthwaite equation approximation of degrees
        of freedom.""",
    )
    confidence_interval: float = Field(
        title="Confidence Interval",
        description="""Confidence interval for relative difference ('diff`)
        of means of this variant and control variant - `[mean - confidence_interval, mean + confidence_interval]`.
        Associated confidence level is the next parameter.""",
    )
    confidence_level: float = Field(
        title="Confidence Level (Statistical Significance)",
        description="""Confidence level used
        to compute (obtain) `confidence_interval`.""",
    )

    @staticmethod
    def from_df(df: pd.DataFrame):
        return [
            MetricStat(
                exp_variant_id=r["exp_variant_id"],
                diff=r["diff"],
                mean=r["mean"],
                sum_value=r["sum_value"],
                p_value=r["p_value"],
                confidence_interval=r["confidence_interval"],
                confidence_level=r["confidence_level"],
            )
            for i, r in df.iterrows()
        ]


class MetricResult(BaseModel):
    """
    Result of single metric evaluation.
    """

    id: int = Field(
        title="Metric Id",
        description="""Database id of the metric, not used at the moment in ep-stats""",
    )
    name: str = Field(
        title="Metric Name",
        description="""Official metric name as it appears in EP.
        The name is only for debugging and has no meaning for ep-stats.""",
    )
    stats: List[MetricStat] = Field(
        title="Per-variant statistics",
        description="""List with one entry per
        variant statistical results.""",
    )

    @staticmethod
    def from_df(metrics: List[Metric], df: pd.DataFrame):
        return [MetricResult(id=m.id, name=m.name, stats=MetricStat.from_df(df[df.metric_id == m.id])) for m in metrics]


class CheckStat(BaseModel):
    variable_id: str = Field(
        title="Check Variable",
        description="""Every check can return different
        variables and their values. E.g. SRM check returns `test_stat` and `p_value` variables with
        their `value`s.""",
    )
    value: float = Field(
        title="Value of the Variable",
        description="""Value of some variable returned by
        the check. E.g. SRM check returns `test_stat` and `p_value` variables with
        their `value`s.""",
    )

    @staticmethod
    def from_df(df: pd.DataFrame):
        return [CheckStat(variable_id=r["variable_id"], value=r["value"]) for i, r in df.iterrows()]


class CheckResult(BaseModel):
    """
    Result of single check evaluation.
    """

    id: int = Field(
        title="Check Id",
        description="Database id of the check, not used at the moment.",
    )
    name: str = Field(
        title="Check Name",
        description="""Official check name as it appears in EP.
        The name is only for debugging and has no meaning for ep-stats.""",
    )
    stats: List[CheckStat] = Field(
        title="Per-variant statistics",
        description="""List with one entry per
        variant statistical results.""",
    )

    @staticmethod
    def from_df(checks: List[Check], df: pd.DataFrame):
        return [CheckResult(id=c.id, name=c.name, stats=CheckStat.from_df(df[df.check_id == c.id])) for c in checks]


class ExposureStat(BaseModel):
    """
    Exposures in the experiment per-variant.
    """

    exp_variant_id: str = Field(title="Variant in the Experiment")
    count: int = Field(
        title="Per-variant exposures",
        description="""Exposure count of experiment (randomization) unit.""",
    )

    @staticmethod
    def from_df(df: pd.DataFrame):
        return [ExposureStat(exp_variant_id=r["exp_variant_id"], count=r["exposures"]) for i, r in df.iterrows()]


class ExposureResult(BaseModel):
    """
    Exposures in the experiment.
    """

    unit_type: str = Field(
        title="Experiment/randomization Unit Type",
        description="""Experiment (randomization) unit type is
    needed to correctly retrieve number of exposures per experiment variant.""",
    )
    stats: List[ExposureStat] = Field(
        title="Experiment Exposures",
        description="""List with experiment variant exposure counts per entry.""",
    )

    @staticmethod
    def from_df(experiment: Experiment, df: pd.DataFrame):
        return ExposureResult(unit_type=experiment.unit_type, stats=ExposureStat.from_df(df))


class Result(BaseModel):
    """
    Result of experiment evaluation.

    Top-level element in the response.
    """

    id: str = Field(
        title="Experiment Id",
    )
    metrics: List[MetricResult] = Field(
        title="Metric Results",
        description="""List with one entry per evaluated metric.""",
    )
    checks: List[CheckResult] = Field(
        title="Check Results",
        description="""List with one entry per evaluated check.""",
    )
    exposure: ExposureResult = Field(title="Experiment Exposures")

    @staticmethod
    def from_evaluation(experiment: Experiment, evaluation: Evaluation):
        metrics = MetricResult.from_df(experiment.metrics, evaluation.metrics)
        checks = CheckResult.from_df(experiment.checks, evaluation.checks)
        exposure = ExposureResult.from_df(experiment, evaluation.exposures)

        return Result(id=experiment.id, metrics=metrics, checks=checks, exposure=exposure)

    class Config:
        schema_extra = {
            "example": {
                "id": "test-conversion",
                "metrics": [
                    {
                        "id": 1,
                        "name": "Click-through Rate",
                        "stats": [
                            {
                                "exp_variant_id": "a",
                                "diff": 0,
                                "mean": 0.23809523809523808,
                                "sum_value": 5,
                                "p_value": 1,
                                "confidence_interval": 1.1432928868841614,
                                "confidence_level": 0.95,
                            },
                            {
                                "exp_variant_id": "b",
                                "diff": 0.13076923076923078,
                                "mean": 0.2692307692307692,
                                "sum_value": 7,
                                "p_value": 1,
                                "confidence_interval": 1.2327467657322932,
                                "confidence_level": 0.95,
                            },
                            {
                                "exp_variant_id": "c",
                                "diff": 0.26,
                                "mean": 0.3,
                                "sum_value": 9,
                                "p_value": 1,
                                "confidence_interval": 1.352808784877644,
                                "confidence_level": 0.95,
                            },
                        ],
                    }
                ],
                "checks": [
                    {
                        "id": 1,
                        "name": "SRM",
                        "stats": [
                            {"variable_id": "p_value", "value": 0.4528439055646014},
                            {"variable_id": "test_stat", "value": 1.5844155844155843},
                            {"variable_id": "confidence_level", "value": 0.999},
                        ],
                    }
                ],
                "exposure": {
                    "unit_type": "test_unit_type",
                    "stats": [
                        {"exp_variant_id": "a", "count": 21},
                        {"exp_variant_id": "b", "count": 26},
                        {"exp_variant_id": "c", "count": 30},
                    ],
                },
            }
        }



from typing import Set
import pandas as pd
import numpy as np

from .parser import Parser


class Metric:
    """
    Definition of a metric to evaluate in an experiment.
    """

    def __init__(self, id: int, name: str, nominator: str, denominator: str):
        """
        Constructor of the general metric definition.

        Parameters `nominator` and `denominator` specify exactly type of data and aggregation of the metric nominator
        and denominator.

        Parameters `format` and `multiplier` does not play any role in metric evaluation. They are used independently
        after the metric evaluation.

        Arguments:
            id: metric (order) id
            name: metric name
            nominator: definition of nominator
            denominator: definition of denominator

        Usage:

        ```python
        Metric(
            1,
            'Click-through Rate',
            'count(test_unit_type.unit.click)',
            'count(test_unit_type.global.exposure)')
        ```
        """

        self.id = id
        self.name = name
        self.nominator = nominator
        self.denominator = denominator
        self._parser = Parser(nominator, denominator)
        self._goals = self._parser.get_goals()

    def get_goals(self) -> Set:
        """
        Get all goals needed to evaluate the metric.
        """
        return self._goals

    def get_evaluate_columns_agg(self, goals: pd.DataFrame) -> np.array:
        """
        Get `count`, `sum_value`, `sum_value_sqr` numpy array of shape (variants, metrics) after
        evaluating nominator and denominator expressions from pre-aggregated goals.

        Arguments:
            goals: one row per experiment variant

        See [`Experiment.evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_agg] for details
        on `goals` at input.

        Returns:
            numpy array of shape (variants, metrics) where metrics are in order of
            (count, sum_value, sum_sqr_value)
        """
        return self._parser.evaluate_agg(goals)

    def get_evaluate_columns_by_unit(self, goals: pd.DataFrame) -> np.array:
        """
        Get `count`, `sum_value`, `sum_value_sqr` numpy array of shape (variants, metrics) after
        evaluating nominator and denominator expressions from goals aggregated by unit.

        Arguments:
            goals: one row per experiment variant

        See [`Experiment.evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_by_unit] for details
        on `goals` at input.

        Returns:
            numpy array of shape (variants, metrics) where metrics are in order of
            (count, sum_value, sum_sqr_value)
        """
        return self._parser.evaluate_by_unit(goals)


class SimpleMetric(Metric):
    """
    Simplified metric definition to evaluate in an axperiment.
    """

    def __init__(
        self,
        id: int,
        name: str,
        numerator: str,
        denominator: str,
        unit_type: str = "test_unit_type",
        metric_format: str = "{:.2%}",
        metric_value_multiplier: int = 1,
    ):
        """
        Constructor of the simplified metric definition.

        It modifies parameters numerator and denominator in a way that it is in line with general Metric definition.
        It adds all the niceties necessary for proper Metric format. Finaly it calls constructor of the parent Metric
        class.

        Arguments:
            id: metric (order) id
            name: metric name
            numerator: value (column) of the numerator
            denominator: value (column) of the denominator
            unit_type: unit type
            metric_format: specify format of the metric, e.g. '${:,.1f}' for RPM
            metric_value_multiplier: specify multiplier, e.g. 1000 for RPM

        Usage:

        ```python
        SimpleMetric(
            1,
            'Click-through Rate',
            'clicks',
            'views',
            unit_type='test_unit_type',
            metric_format='{:.2%}',
            metric_value_multiplier=1)
        ```
        """
        agg_type = "global"  # technical parameter; it has no impact
        num = "value" + "(" + unit_type + "." + agg_type + "." + numerator + ")"
        den = "value" + "(" + unit_type + "." + agg_type + "." + denominator + ")"

        super().__init__(id, name, num, den)

        self.metric_format = metric_format
        self.metric_value_multiplier = metric_value_multiplier



from typing import List, Optional, Any
from pydantic import BaseModel, validator, root_validator, Field
from pyparsing import ParseException
from datetime import datetime
from statsd import StatsClient

from ..toolkit import Experiment as EvExperiment, Filter as EvFilter, FilterScope
from ..toolkit import Metric as EvMetric
from ..toolkit import SrmCheck as EvSrmCheck
from ..toolkit import Parser


class Metric(BaseModel):
    """
    Defines metric to evaluate.
    """

    id: int = Field(
        title="Metric Id",
        description="""Database id of the metric, not used at the moment in ep-stats. We only
        repeat this id in response so the caller knows what result belongs to what request.""",
    )
    name: str = Field(
        title="Metric Name",
        description="""Official metric name as it appears in EP.
        The name is only for debugging and has no meaning for ep-stats. We only repeat the name in the response
        for caller's convenience.""",
    )
    nominator: str = Field(
        title="Metric Nominator",
        description="""EP metric is defined in the form of `nominator / denominator`.
        Both parts are entered as expressions. Example: `count(my_unit_type.unit.conversion)`.""",
    )
    denominator: str = Field(
        title="Metric Denominator",
        description="""EP metric is defined in the form of `nominator / denominator`.
        Both parts are entered as expressions. Example: `count(my_unit_type.unit.conversion)`.""",
    )

    @validator("id")
    def id_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect id to be non-empty")
        return value

    @validator("name")
    def name_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect name to be non-empty")
        return value

    @validator("nominator")
    def nominator_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect nominator to be non-empty")
        return value

    @validator("denominator")
    def denominator_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect denominator to be non-empty")
        return value

    @root_validator
    def check_nominator_denominator(cls, values):
        nominator, denominator = values.get("nominator"), values.get("denominator")
        if not nominator:
            raise ValueError("we expect nominator to be non-empty")
        if not denominator:
            raise ValueError("we expect denominator to be non-empty")
        try:
            parser = Parser(nominator, denominator)
            if not parser.get_goals():
                raise ValueError("We expect the metric to have at least one goal in nominator and denominator")
            return values
        except ParseException as e:
            raise ValueError(f"Cannot parse nominator '{nominator}' or '{denominator}' because of '{e}'")

    def to_metric(self):
        return EvMetric(self.id, self.name, self.nominator, self.denominator)


class Check(BaseModel):
    """
    Defines metric to evaluate.

    Only `SRM` check is available at the moment.
    """

    id: int = Field(
        title="Check Id",
        description="""Database id of the check, not used at the moment in ep-stats. We only
        repeat this id in response so the caller knows what result belongs to what request.""",
    )
    name: str = Field(
        title="Check Name",
        description="""Official check name as it appears in EP. The name is only for debugging
        and has no meaning for ep-stats. We only repeat the name in the response for caller's convenience.""",
    )
    denominator: str = Field(
        title="Check Denominator",
        description="""EP metric is defined in the form of `nominator / denominator`.
        Both parts are entered as expressions.

        The only supported check is SRM check and it should be called with denominator
        of `count(my_unit_type.global.exposure)`.""",
    )

    @validator("id")
    def id_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect id to be non-empty")
        return value

    @validator("name")
    def name_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect name to be non-empty")
        return value

    @validator("denominator")
    def denominator_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect denominator to be non-empty")
        return value

    @validator("denominator")
    def check_denominator(cls, value):
        try:
            parser = Parser(value, value)
            if not parser.get_goals():
                raise ValueError("We expect the check to have at least one goal in denominator")
            return value
        except ParseException as e:
            raise ValueError(f"Cannot parse '{value}' because of '{e}'")

    def to_check(self):
        return EvSrmCheck(self.id, self.name, self.denominator)


class Filter(BaseModel):
    """
    Filter specification for data to evaluate.
    """

    dimension: str = Field(title="Name of the dimension")

    value: List[Any] = Field(title="List of possible values")

    scope: FilterScope = Field(
        title="Scope of the filter", description="Scope of the filter is either `exposure` or `goal`."
    )

    @validator("dimension")
    def dimension_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect dimension to be non-empty")
        return value

    @validator("scope")
    def scope_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect scope to be either `exposure` or `goal`")
        return value

    @validator("value")
    def value_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect value to be non-empty")
        return value

    def to_filter(self):
        return EvFilter(self.dimension, self.value, self.scope)


class Experiment(BaseModel):
    """
    Experiment to evaluate.
    """

    id: str = Field(
        title="Experiment Id",
    )

    control_variant: str = Field(
        title="Identification of the control variant",
        description="""All other variant data
        in the experiment are evaluated against this control variant.""",
    )

    variants: Optional[List[str]] = Field(
        title="Variants", description="""List of experiment variants to evaluate for."""
    )

    unit_type: str = Field(
        title="Experiment/randomization Unit Type",
        description="""Experiment (randomization) unit type is
        needed to correctly retrieve number of exposures per experiment variant.""",
    )

    date_from: Optional[str] = Field(
        title="Start of the date range",
        description="""Required format is `2020-06-15`.
        `date_from` is optional, data are unbound from the start when `date_from` is not present.""",
        default=None,
    )

    date_to: Optional[str] = Field(
        title="End of the date range (inclusive)",
        description="""Required format is `2020-06-15`.
        `date_to` is optional, data are unbound from the end when `date_to` is not present.""",
        default=None,
    )

    date_for: Optional[str] = Field(
        title="Date of the experiment to evaluate for",
        description="""Required format is `2020-06-15`.
        `date_for` is optional. If present, both `date_from` and `date_to` must be present too and `date_for` must be
        between `date_from` and `date_to`. `date_from` and `date_to` set expected experiment duration,
        while `date_for` informs about position within the experiment duration. It is used for sequential analysis.""",
        default=None,
    )
    metrics: List[Metric] = Field(title="List of metrics to evaluate")

    checks: List[Check] = Field(title="List of checks to evaluate")

    filters: Optional[List[Filter]] = Field(
        title="Filtering conditions", description="""List of filtering conditions to apply on exposure and goals."""
    )

    @validator("id")
    def id_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect id to be non-empty")
        return value

    @validator("control_variant")
    def control_variant_must_be_not_empty(cls, value):
        if not value:
            raise ValueError("we expect control_variant to be non-empty")
        return value

    @validator("date_from")
    def date_from_must_be_date(cls, value):
        if value is not None:
            try:
                datetime.strptime(value, "%Y-%m-%d")
            except ValueError:
                raise ValueError("we expect date_from to be in `2020-06-15` format")

        return value

    @validator("date_to")
    def date_to_must_be_date(cls, value):
        if value is not None:
            try:
                datetime.strptime(value, "%Y-%m-%d")
            except ValueError:
                raise ValueError("we expect date_to to be in `2020-06-15` format")

        return value

    @root_validator
    def check_date_from_to(cls, values):
        date_from, date_to, date_for = (
            values.get("date_from"),
            values.get("date_to"),
            values.get("date_for"),
        )
        if date_for is not None and (date_from is None or date_to is None):
            raise ValueError("date_for requires date_from and date_to to be present as well")
        if date_from is not None and date_to is not None:
            try:
                df = datetime.strptime(date_from, "%Y-%m-%d")
                dt = datetime.strptime(date_to, "%Y-%m-%d")
            except ValueError:
                raise ValueError("cannot parse date_from, date_to")
            if date_for is not None:
                try:
                    dfor = datetime.strptime(date_for, "%Y-%m-%d")
                except ValueError:
                    raise ValueError("cannot parse date_for")
                if dfor < df:
                    raise ValueError("we expect date_for to be greater or equal to date_from")
                if dfor > dt:
                    raise ValueError("we expect date_for to be less or equal to date_to")
            if dt < df:
                raise ValueError("we expect date_to to be greater or equal to date_from")

        return values

    def to_experiment(self, statsd: StatsClient):
        metrics = [m.to_metric() for m in self.metrics]
        checks = [c.to_check() for c in self.checks]
        return EvExperiment(
            self.id,
            self.control_variant,
            metrics,
            checks,
            date_from=self.date_from,
            date_to=self.date_to,
            date_for=self.date_for,
            unit_type=self.unit_type,
            variants=self.variants,
            statsd=statsd,
            filters=[f.to_filter() for f in self.filters] if self.filters else [],
        )

    class Config:
        schema_extra = {
            "example": {
                "id": "test-conversion",
                "variants": ["a", "b", "c"],
                "control_variant": "a",
                "unit_type": "test_unit_type",
                "filters": [
                    {"dimension": "element", "value": ["button-1"], "scope": "goal"},
                    {"dimension": "browser", "value": ["firefox"], "scope": "exposure"},
                ],
                "metrics": [
                    {
                        "id": 1,
                        "name": "Click-through Rate",
                        "nominator": "count(test_unit_type.unit.click)",
                        "denominator": "count(test_unit_type.global.exposure)",
                    }
                ],
                "checks": [
                    {
                        "id": 1,
                        "name": "SRM",
                        "denominator": "count(test_unit_type.global.exposure)",
                    }
                ],
            }
        }



import logging
from concurrent.futures import ThreadPoolExecutor
from statsd import StatsClient
from pydantic import BaseSettings

from .toolkit.testing import TestDaoFactory, TestData
from .server import get_api, serve, ApiSettings


class Settings(BaseSettings):
    api: ApiSettings = ApiSettings()

    statsd_host: str = "localhost"
    statsd_port: int = 8888

    evaluation_pool_size: int = 5


settings = Settings()


def get_dao_factory():
    return TestDaoFactory(TestData())


def get_dao():
    try:
        dao = get_dao_factory().get_dao()
        yield dao
    finally:
        dao.close()


def get_executor_pool():
    try:
        evaluation_pool_size = settings.evaluation_pool_size
        evaluation_pool = ThreadPoolExecutor(max_workers=evaluation_pool_size)
        yield evaluation_pool
    finally:
        pass


def get_statsd():
    try:
        prefix = f"{settings.api.app_name}.{settings.api.app_env}"
        statsd = StatsClient(settings.statsd_host, settings.statsd_port, prefix=prefix)
        yield statsd
    finally:
        pass


api = get_api(settings.api, get_dao, get_executor_pool, get_statsd)


def main():
    from .config import config

    logging.config.dictConfig(config["logging"])
    serve("epstats:api", settings.api, config["logging"])


if __name__ == "__main__":
    main()



import logging
from fastapi import APIRouter, Depends, HTTPException
from statsd import StatsClient
import asyncio
from concurrent.futures import ThreadPoolExecutor

from ..toolkit import Experiment as EvExperiment
from ..toolkit import Dao

from .req import Experiment
from .res import Result


_logger = logging.getLogger("epstats")


def get_evaluate_router(get_dao, get_executor_pool, get_statsd) -> APIRouter:
    def _evaluate(experiment: EvExperiment, dao: Dao, statsd: StatsClient):
        try:
            with statsd.timer("timing.evaluation"):
                _logger.info(f"Evaluating experiment [{experiment.id}]")
                _logger.debug(f"Loading goals for experiment [{experiment.id}]")
                with statsd.timer("timing.query"):
                    goals = dao.get_agg_goals(experiment).sort_values(["exp_variant_id", "goal"])
                    _logger.info(f"Retrieved {len(goals)} goals in experiment [{experiment.id}]")
                with statsd.timer("timing.stats"):
                    evaluation = experiment.evaluate_agg(goals)
                    statsd.incr("evaluations")
                _logger.info(
                    (
                        f"Evaluation of experiment [{experiment.id}] finished with evaluation"
                        f" of {evaluation.metrics.metric_id.nunique()} "
                        f"metrics and {evaluation.checks.check_id.nunique()} checks"
                    )
                )
            return Result.from_evaluation(experiment, evaluation)
        except Exception as e:
            _logger.error(f"Cannot evaluate experiment [{experiment.id}] because of {e}")
            _logger.exception(e)
            statsd.incr("errors.experiment")
            raise HTTPException(
                status_code=500,
                detail=f"Cannot evaluate experiment [{experiment.id}] because of {e}",
            )

    router = APIRouter()

    @router.post("/evaluate", response_model=Result, tags=["Experiment Evaluation"])
    async def evaluate_experiment(
        experiment: Experiment,
        evaluation_pool: ThreadPoolExecutor = Depends(get_executor_pool),
        dao: Dao = Depends(get_dao),
        statsd: StatsClient = Depends(get_statsd),
    ):
        """
        Evaluates single `Experiment`.
        """
        _logger.info(f"Calling evaluate with {experiment.json()}")
        statsd.incr("requests.evaluate")
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(evaluation_pool, _evaluate, experiment.to_experiment(statsd), dao, statsd)

    return router



config = {
    "logging": {
        "version": 1,
        "formatters": {
            "standard": {
                "class": "logging.Formatter",
                "format": "%(asctime)s [%(levelname)-7s] %(name)-35s %(message)s",
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "epstats": {
                "level": "INFO",
                "handlers": ["console"],
                "propagate": False,
            },
        },
        "root": {
            "level": "INFO",
            "handlers": ["console"],
        },
    },
}



import logging
from typing import List, Any
from enum import Enum
import pandas as pd
import numpy as np
from datetime import datetime
from statsd import StatsClient
from dataclasses import dataclass

from .metric import Metric
from .check import Check
from .utils import get_utc_timestamp, goals_wide_to_long
from .parser import EpGoal, UnitType, AggType, Goal

from .statistics import Statistics


class Evaluation:
    """
    Results of an experiment evaluation.
    """

    def __init__(self, metrics: pd.DataFrame, checks: pd.DataFrame, exposures: pd.DataFrame):
        self.metrics = metrics
        self.checks = checks
        self.exposures = exposures

    @classmethod
    def metric_columns(cls) -> List[str]:
        """
        `metrics` dataframe with columns:

        1. `timestamp` - timestamp of evaluation
        1. `exp_id` - experiment id
        1. `metric_id` - metric id as in [`Experiment`][epstats.toolkit.experiment.Experiment] definition
        1. `metric_name` - metric name as in [`Experiment`][epstats.toolkit.experiment.Experiment] definition
        1. `exp_variant_id` - variant id
        1. `count` - number of exposures, value of metric denominator
        1. `mean` - `sum_value` / `count`
        1. `std` - sample standard deviation
        1. `sum_value` - value of goals, value of metric nominator
        1. `confidence_level` - current confidence level used to calculate `p_value` and `confidence_interval`
        1. `diff` - relative diff between sample means of this and control variant
        1. `test_stat` - value of test statistic of the relative difference in means
        1. `p_value` - p-value of the test statistic under current `confidence_level`
        1. `confidence_interval` - confidence interval of the `diff` under current `confidence_level`
        1. `standard_error` - standard error of the `diff`
        1. `degrees_of_freedom` - degrees of freedom of this variant mean
        """
        return [
            "timestamp",
            "exp_id",
            "metric_id",
            "metric_name",
            "exp_variant_id",
            "count",
            "mean",
            "std",
            "sum_value",
            "confidence_level",
            "diff",
            "test_stat",
            "p_value",
            "confidence_interval",
            "standard_error",
            "degrees_of_freedom",
        ]

    @classmethod
    def check_columns(cls) -> List[str]:
        """
        `checks` dataframe with columns:

        1. `timestamp` - timestamp of evaluation
        1. `exp_id` - experiment id
        1. `check_id` - check id as in [`Experiment`][epstats.toolkit.experiment.Experiment] definition
        1. `variable_id` - name of the variable in check evaluation, SRM check has following variables `p_value`,
        `test_stat`, `confidence_level`
        1. `value` - value of the variable
        """
        return ["timestamp", "exp_id", "check_id", "check_name", "variable_id", "value"]

    @classmethod
    def exposure_columns(cls) -> List[str]:
        """
        `exposures` dataframe with columns:

        1. `timestamp` - timestamp of evaluation
        1. `exp_id` - experiment id
        1. `exp_variant_id` - variant id
        1. `exposures` - number of exposures of this variant
        """
        return ["exp_variant_id", "exposures"]


class FilterScope(str, Enum):
    """
    Scope of data where to apply the filter.
    """

    exposure = "exposure"
    goal = "goal"


@dataclass
class Filter:
    """
    Filter specification for data to evaluate.
    """

    dimension: str
    value: List[Any]
    scope: FilterScope


class Experiment:
    """
    Evaluate one experiment described as a list of metrics and checks.

    See [Statistics](../stats/basics.md) for details about statistical method used
    and [`Evaluation`][epstats.toolkit.experiment.Evaluation] for description of output.
    """

    def __init__(
        self,
        id: str,
        control_variant: str,
        metrics: List[Metric],
        checks: List[Check],
        unit_type: str,
        date_from: str = None,
        date_to: str = None,
        date_for: str = None,
        confidence_level: float = 0.95,
        variants: List[str] = None,
        statsd: StatsClient = StatsClient(),
        filters: List[Filter] = None,
    ):
        self._logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        self.id = id
        self.control_variant = control_variant
        self.unit_type = unit_type
        self.metrics = metrics
        self.checks = checks
        self.date_from = datetime.strptime(date_from, "%Y-%m-%d").date() if date_from is not None else None
        self.date_to = datetime.strptime(date_to, "%Y-%m-%d").date() if date_to is not None else None
        self.date_for = (
            datetime.strptime(date_for, "%Y-%m-%d").date() if date_for is not None else datetime.today().date()
        )
        self.confidence_level = confidence_level
        self.variants = variants
        self._exposure_goals = [
            EpGoal(
                [
                    "count",
                    "(",
                    UnitType([unit_type]),
                    ".",
                    AggType(["global"]),
                    ".",
                    Goal(["exposure"]),
                    ")",
                ]
            )
        ]
        self.statsd = statsd
        self.filters = filters if filters is not None else []

    def evaluate_agg(self, goals: pd.DataFrame) -> Evaluation:
        """
        Evaluate all metrics and checks in the experiment from already pre-aggregated goals.

        This method is usefull when there are too many units in the experiment to evaluate it
        using [`evaluate_by_unit`][epstats.toolkit.experiment.Experiment.evaluate_by_unit].

        Does best effort to fill in missing goals and variants with zeros.

        Arguments:
            goals: dataframe with one row per goal and aggregated data in columns

        `goals` dataframe columns:

        1. `exp_id` - experiment id
        1. `exp_variant_id` - variant
        1. `unit_type` - randomization unit type
        1. `agg_type` - level of aggregation
        1. `goal` - goal name
        1. `dimension` - name of the dimension, e.g. `product`
        1. `dimension_value` - value of the dimension, e.g. `p_1`
        1. `count` - number of observed goals (e.g. conversions)
        1. `sum_sqr_count` - summed squared number of observed goals per unit, it is similar
        to `sum_sqr_value`
        1. `sum_value` - value of observed goals
        1. `sum_sqr_value` - summed squared value per unit.  This is used to calculate
        sample standard deviation from pre-aggregated data (it is a term $\sum x^2$
        in $\hat{\sigma}^2 = \\frac{\sum x^2 - \\frac{(\sum x)^2}{n}}{n-1}$).
        1. `count_unique` - number of units with at least 1 observed goal

        Returns:
            set of dataframes with evaluation

        Usage:

        ```python
        from epstats.toolkit import Experiment, Metric, SrmCheck
        experiment = Experiment(
            'test-conversion',
            'a',
            [Metric(
                1,
                'Click-through Rate',
                'count(test_unit_type.unit.click)',
                'count(test_unit_type.global.exposure)'),
            ],
            [SrmCheck(1, 'SRM', 'count(test_unit_type.global.exposure)')],
            unit_type='test_unit_type')

        # This gets testing data, use other Dao or get aggregated goals in some other way.
        from epstats.toolkit.testing import TestData
        goals = TestData.load_goals_agg(experiment.id)

        # evaluate experiment
        ev = experiment.evaluate_agg(goals)

        # work with results
        print(ev.exposures)
        print(ev.metrics[ev.metrics == 1])
        print(ev.checks[ev.checks == 1])

        # this is to assert that this code sample works correctly
        from epstats.toolkit.testing import TestDao
        assert_experiment(experiment, ev, TestDao(TestData()))
        ```

        Input data frame example:

        ```
        exp_id      exp_variant_id  unit_type           agg_type    goal            dimension       dimension_value     count   sum_sqr_count   sum_value   sum_sqr_value   count_unique
        test-srm    a               test_unit_type      global      exposure                                            100000  100000          100000      100000          100000
        test-srm    b               test_unit_type      global      exposure                                            100100  100100          100100      100100          100100
        test-srm    a               test_unit_type      unit        conversion                                            1200    1800            32000       66528           900
        test-srm    a               test_unit_type_2    global      conversion      product         product_1           1000    1700            31000       55000           850
        ```
        """
        g = self._fix_missing_agg(goals)
        return self._evaluate(
            g,
            Experiment._metrics_column_fce_agg,
            Experiment._checks_fce_agg,
            Experiment._exposures_fce_agg,
        )

    def evaluate_wide_agg(self, goals: pd.DataFrame) -> Evaluation:
        """
        This is a simplified version of the method [`evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_agg].

        It consumes simple input `goals` dataframe, transfers it into suitable dataframe format and evaluate it using general method [`evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_agg].

        It assumes that the first two columns are name of the experiment and variants. Than follows columns with data.

        Arguments:
            goals: dataframe with one row per variant and aggregated data in columns

        Possible `goals` dataframe columns (check the input dataframe example):

        1. `exp_id` - experiment id
        1. `exp_variant_id` - variant
        1. `clicks` - sum of clicks
        1. `views` - sum of views
        1. `bookings` - sum of bookings
        1. `bookings_squared` - sum of bookings squared

        Returns:
            set of dataframes with evaluation

        Usage:

        ```python
        from epstats.toolkit import Experiment, SimpleMetric, SimpleSrmCheck
        from epstats.toolkit.results import results_long_to_wide, format_results
        from epstats.toolkit.testing import TestData

        # Load Test Data
        goals = TestData.load_goals_simple_agg()

        # Define the experiment
        unit_type = 'test_unit_type'
        experiment = Experiment(
            'my-experiment',
            'a',
            [
                SimpleMetric(1, 'Click-through Rate (CTR)', 'clicks', 'views', unit_type),
                SimpleMetric(2, 'Conversion Rate', 'conversions', 'views', unit_type),
                SimpleMetric(3, 'Revenue per Mille (RPM)', 'bookings', 'views', unit_type, metric_format='${:,.2f}', metric_value_multiplier=1000),
            ],
            [SimpleSrmCheck(1, 'SRM', 'views')],
            unit_type=unit_type)

        # Evaluate the experiment
        ev = experiment.evaluate_wide_agg(goals)

        # Work with results
        print(ev.exposures)
        print(ev.metrics)
        print(ev.checks)

        # Possible formatting of metrics
        ev.metrics.pipe(results_long_to_wide).pipe(format_results, experiment, format_pct='{:.1%}', format_pval='{:.3f}')
        ```

        Input dataframe example:
        ```
        experiment_id   variant_id  views   clicks  conversions     bookings    bookings_squared
        my-exp          a           473661  48194   413             17152       803105
        my-exp          b           471485  47184   360             14503       677178
        my-exp          c           477159  48841   406             15892       711661
        my-exp          d           474934  49090   289             11995       566700
        ```
        """
        g = goals_wide_to_long(goals)
        return self.evaluate_agg(g)

    def evaluate_by_unit(self, goals: pd.DataFrame) -> Evaluation:
        """
        Evaluate all metrics and checks in the experiment from goals grouped by `unit_id`.

        This method is usefull when there are not many (<1M) units in the experiment to evaluate it.
        If there many units exposed to the experiment, pre-aggregate data and use [`evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_agg].

        Does best effort to fill in missing goals and variants with zeros.

        Arguments:
            goals: dataframe with one row per goal and aggregated data in columns

        `goals` dataframe columns:

        1. `exp_id` - experiment id
        1. `exp_variant_id` - variant
        1. `unit_type` - randomization unit type
        1. `unit_id` - (randomization) unit id
        1. `agg_type` - level of aggregation
        1. `goal` - goal name
        1. `dimension` - name of the dimension, e.g. `product`
        1. `dimension_value` - value of the dimension, e.g. `p_1`
        1. `count` - number of observed goals
        1. `sum_value` - value of observed goals

        Returns:
            set of dataframes with evaluation

        Usage:

        ```python
        from epstats.toolkit import Experiment, Metric, SrmCheck
        experiment = Experiment(
            'test-real-valued',
            'a',
            [Metric(
                2,
                'Average Bookings',
                'value(test_unit_type.unit.conversion)',
                'count(test_unit_type.unit.exposure)')
            ],
            [],
            unit_type='test_unit_type')

        # This gets testing data, use other Dao or get aggregated goals in some other way.
        from epstats.toolkit.testing import TestData
        goals = TestData.load_goals_by_unit(experiment.id)

        # evaluate experiment
        ev = experiment.evaluate_by_unit(goals)

        # work with results
        print(ev.exposures)
        print(ev.metrics[ev.metrics == 1])
        print(ev.checks[ev.checks == 1])

        # this is to assert that this code sample works correctly
        from epstats.toolkit.testing import TestDao
        assert_experiment(experiment, ev, TestDao(TestData()))
        ```

        Input data frame example:

        ```
        exp_id      exp_variant_id  unit_type       unit_id             agg_type    goal              dimension         dimension_value     count   sum_value
        test-srm    a               test_unit_type  test_unit_type_1    unit        exposure                                                1       1
        test-srm    a               test_unit_type  test_unit_type_1    unit        conversion        product           product_1           2       75
        test-srm    b               test_unit_type  test_unit_type_2    unit        exposure                                                1       1
        test-srm    b               test_unit_type  test_unit_type_3    unit        exposure                                                1       1
        test-srm    b               test_unit_type  test_unit_type_3    unit        conversion        product           product_2           1       1
        ```
        """
        g = self._fix_missing_by_unit(goals)

        # We need to pivot table to get all goals per `unit_id` on the same row in the data frame.
        # This is needed to be able to vector-evaluate compound metrics
        # eg. `value(test_unit_type.unit.conversion) - value(test_unit_type.unit.refund)`
        g = (
            pd.pivot_table(
                g,
                values=["count", "sum_value"],
                index=[
                    "exp_id",
                    "exp_variant_id",
                    "unit_type",
                    "agg_type",
                    "unit_id",
                    "dimension",
                    "dimension_value",
                ],
                columns="goal",
                aggfunc=np.sum,
                fill_value=0,
            )
            .swaplevel(axis=1)
            .reset_index()
        )

        return self._evaluate(
            g,
            Experiment._metrics_column_fce_by_unit,
            Experiment._checks_fce_by_unit,
            Experiment._exposures_fce_by_unit,
        )

    def get_goals(self) -> List[EpGoal]:
        """
        List of all goals needed to evaluate all metrics and checks in the experiment.

        Returns:
            list of parsed structured goals
        """
        res = set()
        for m in self.metrics:
            res = res.union(m.get_goals())
        for c in self.checks:
            res = res.union(c.get_goals())
        res = res.union(self._exposure_goals)
        return list(res)

    @staticmethod
    def _metrics_column_fce_agg(m: Metric, goals: pd.DataFrame):
        """
        Gets count, sum_value, sum_sqr_value columns by expression from already aggregated goals.
        """
        return m.get_evaluate_columns_agg(goals)

    @staticmethod
    def _metrics_column_fce_by_unit(m: Metric, goals: pd.DataFrame):
        """
        Gets count, sum_value, sum_sqr_value columns by expression from goals grouped by `unit_id`.
        """
        return m.get_evaluate_columns_by_unit(goals)

    @staticmethod
    def _checks_fce_agg(c: Check, goals: pd.DataFrame, control_variant: str):
        """
        Evaluates checks from already aggregated goals.
        """
        return c.evaluate_agg(goals, control_variant)

    @staticmethod
    def _checks_fce_by_unit(c: Check, goals: pd.DataFrame, control_variant: str):
        """
        Evaluates checks from goals grouped by `unit_id`.
        """
        return c.evaluate_by_unit(goals, control_variant)

    @staticmethod
    def _exposures_fce_agg(goals: pd.DataFrame, exp_id: str, unit_type: str):
        """
        Evaluates checks from already aggregated goals.
        """
        df = (
            goals[(goals["unit_type"] == unit_type) & (goals["agg_type"] == "global") & (goals["goal"] == "exposure")]
            .groupby("exp_variant_id")
            .agg(exposures=("count", "sum"))
            .reset_index()
        )
        df["exp_id"] = exp_id
        return df

    @staticmethod
    def _exposures_fce_by_unit(goals: pd.DataFrame, exp_id: str, unit_type: str):
        """
        Evaluates checks from already aggregated goals.
        """
        df = goals[(goals["unit_type"] == unit_type) & (goals["agg_type"] == "unit")][
            [("exp_variant_id", ""), ("exposure", "count")]
        ]
        df = df.droplevel(0, axis=1)
        df.columns = ["exp_variant_id", "exposures"]
        d = df.groupby("exp_variant_id").agg(exposures=("exposures", "sum")).reset_index()
        d["exp_id"] = exp_id
        return d

    def _evaluate(self, goals: pd.DataFrame, metrics_column_fce, checks_fce, exposures_fce):
        metrics = self._evaluate_metrics(goals, metrics_column_fce)
        checks = self._evaluate_checks(goals, checks_fce)
        exposures = self._evaluate_exposures(goals, exposures_fce)
        return Evaluation(metrics, checks, exposures)

    def _evaluate_exposures(self, goals: pd.DataFrame, exposures_fce) -> pd.DataFrame:
        return exposures_fce(goals, self.id, self.unit_type)

    def _evaluate_checks(self, goals: pd.DataFrame, check_fce) -> pd.DataFrame:
        res = []
        for c in self.checks:
            try:
                r = check_fce(c, goals, self.control_variant)
                r["exp_id"] = self.id
                res.append(r)
            except Exception as e:
                self._logger.warning(f"Cannot evaluate check [{c.id} in experiment [{self.id}] because of {e}")
                self.statsd.incr("errors.check")

        c = pd.concat(res, axis=1) if res != [] else pd.DataFrame([], columns=Evaluation.check_columns())
        c["timestamp"] = round(get_utc_timestamp(datetime.now()).timestamp())
        return c[Evaluation.check_columns()]

    def _fix_missing_agg(self, goals: pd.DataFrame) -> pd.DataFrame:
        """
        Adds zero values for missing goals and variants that are needed for metric evaluation.

        Does that in the best effort - fills in `count`, `sum_sqr_count`, `sum_value`, `sum_sqr_value` and `count_unique` with zeros.
        """
        # what variants and goals there should be from all the goals needed to evaluate all metrics
        self.variants = (
            self.variants
            if self.variants is not None
            else np.unique(np.append(goals["exp_variant_id"], self.control_variant))
        )
        g = goals[goals.exp_variant_id.isin(self.variants)]
        nvs = self.variants
        ngs = self.get_goals()

        # variants * goals is the number of variant x goals combinations we expect in the data
        lnvs = len(nvs)
        lngs = len(ngs)
        ln = lnvs * lngs

        # create zero data frame for all variants and goals
        empty_df = pd.DataFrame(
            {
                "exp_variant_id": np.tile(nvs, lngs),
                "unit_type": np.repeat([g.unit_type for g in ngs], lnvs),
                "agg_type": np.repeat([g.agg_type for g in ngs], lnvs),
                "goal": np.repeat([g.goal for g in ngs], lnvs),
                "dimension": np.repeat([g.dimension for g in ngs], lnvs),
                "dimension_value": np.repeat([g.dimension_value for g in ngs], lnvs),
                "count": np.zeros(ln),
                "sum_sqr_count": np.zeros(ln),
                "sum_value": np.zeros(ln),
                "sum_sqr_value": np.zeros(ln),
                "count_unique": np.zeros(ln),
            }
        )

        # join to existing data and use zeros for only missing variants and goals
        m = (
            pd.concat([g, empty_df], axis=0)
            .groupby(
                [
                    "exp_variant_id",
                    "unit_type",
                    "agg_type",
                    "dimension",
                    "dimension_value",
                    "goal",
                ]
            )
            .sum()
            .reset_index()
        )
        return m

    def _fix_missing_by_unit(self, goals: pd.DataFrame) -> pd.DataFrame:
        """
        Adds zero values for missing goals and variants that are needed for metric evaluation.

        Does that in the best effort - fills in `count`, `sum_value` with zeros.
        """
        # what variants and goals there should be from all the goals needed to evaluate all metrics
        self.variants = (
            self.variants
            if self.variants is not None
            else np.unique(np.append(goals["exp_variant_id"], self.control_variant))
        )
        g = goals[goals.exp_variant_id.isin(self.variants)]
        nvs = self.variants
        ngs = self.get_goals()

        # variants * goals is the number of variant x goals combinations we expect in the data
        lnvs = len(nvs)
        lngs = len(ngs)
        ln = lnvs * lngs

        # create zero data frame for all variants and goals
        empty_df = pd.DataFrame(
            {
                "exp_id": np.repeat(self.id, ln),
                "exp_variant_id": np.tile(nvs, lngs),
                "unit_type": np.repeat([g.unit_type for g in ngs], lnvs),
                "agg_type": np.repeat([g.agg_type for g in ngs], lnvs),
                "goal": np.repeat([g.goal for g in ngs], lnvs),
                "dimension": np.repeat([g.dimension for g in ngs], lnvs),
                "dimension_value": np.repeat([g.dimension_value for g in ngs], lnvs),
                "unit_id": np.repeat("fillna", ln),
                "count": np.zeros(ln),
                "sum_value": np.zeros(ln),
            }
        )

        # join to existing data and use zeros for only missing variants and goals
        m = pd.concat([g, empty_df], axis=0)
        return m[
            [
                "exp_id",
                "exp_variant_id",
                "unit_type",
                "agg_type",
                "dimension",
                "dimension_value",
                "goal",
                "unit_id",
                "count",
                "sum_value",
            ]
        ]

    def _evaluate_metrics(self, goals: pd.DataFrame, column_fce) -> pd.DataFrame:
        if not self.metrics:
            return pd.DataFrame([], columns=Evaluation.metric_columns())

        sts = []
        for m in self.metrics:
            count, sum_value, sum_sqr_value = column_fce(m, goals)
            sts.append([count, sum_value, sum_sqr_value])
        stats = np.array(sts).transpose(0, 2, 1)
        metrics = stats.shape[0]
        variants = stats.shape[1]

        count = stats[:, :, 0]
        sum_value = stats[:, :, 1]
        sum_sqr_value = stats[:, :, 2]
        with np.errstate(divide="ignore", invalid="ignore"):
            # We fill in zeros, when goal data are missing for some variant.
            # There could be division by zero here which is expected as we return
            # nan or inf values to the caller.
            mean = sum_value / count
            std = np.sqrt((sum_sqr_value - sum_value * sum_value / count) / (count - 1))

        # sequential testing correction
        if self.date_from is not None and self.date_to is not None:
            # Parameters
            test_length = (self.date_to - self.date_from).days + 1  # test length in days
            actual_day = (self.date_for - self.date_from).days + 1  # day(s) since beginning of the test
            actual_day = min(actual_day, test_length)  # actual day of evaluation must be in interval [1, test_length]

            # confidence level adjustment - applied when actual_day < test_length (test is still running)
            confidence_level = Statistics.obf_alpha_spending_function(self.confidence_level, test_length, actual_day)
        else:
            confidence_level = self.confidence_level  # no change

        stats = np.dstack((count, mean, std, sum_value, np.ones(count.shape) * confidence_level))
        stats = np.dstack(
            (
                np.repeat([m.id for m in self.metrics], variants).reshape(metrics, variants, -1),
                np.repeat([m.name for m in self.metrics], variants).reshape(metrics, variants, -1),
                np.tile(goals["exp_variant_id"].unique(), metrics).reshape(metrics, variants, -1),
                stats,
            )
        )

        # dimensions of `stats` array: (metrics, variants, stats)
        # elements of `stats` array: metrics_id, exp_variant_id, count, mean, std, sum_value, confidence_level
        # hypothesis evaluation (standard way using t-test)
        c = Statistics.ttest_evaluation(stats)

        # multiple variants (comparisons) correction - applied when we have multiple treatment variants
        if variants > 2:
            c = Statistics.multiple_comparisons_correction(c, variants, metrics, confidence_level)

        c["exp_id"] = self.id
        c["timestamp"] = round(get_utc_timestamp(datetime.now()).timestamp())
        return c[Evaluation.metric_columns()]



import pandas as pd
import numpy as np
import scipy.stats as st
from statsmodels.stats.multitest import multipletests
import warnings


class Statistics:
    """
    Various methods needed to evaluate experiment.
    """

    @classmethod
    def ttest_evaluation(cls, stats: np.array) -> pd.DataFrame:
        """
        Testing statistical significance of relative difference in means of treatment and control variant.

        This is inspired by [scipy.stats.ttest_ind_from_stats](https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_ind_from_stats.html)
        method that returns many more statistics than p-value and test statistic.

        Statistics used:

        1. [Welch's t-test](https://en.wikipedia.org/wiki/Welch%27s_t-test)
        1. [Welch–Satterthwaite equation](https://en.wikipedia.org/wiki/Welch%E2%80%93Satterthwaite_equation)
        approximation of degrees of freedom.

        Arguments:
            stats: array with dimensions (metrics, variants, stats)

        `stats` array values:

        0. `metric_id`
        1. `metric_name`
        1. `exp_variant_id`
        1. `count`
        1. `mean`
        1. `std`
        1. `sum_value`
        1. `sum_sqr_value`

        Returns:
            dataframe containing statistics from the t-test

        Schema of returned dataframe:

        1. `metric_id` - metric id as in [`Experiment`][epstats.toolkit.experiment.Experiment] definition
        1. `metric_name` - metric name as in [`Experiment`][epstats.toolkit.experiment.Experiment] definition
        1. `exp_variant_id` - variant id
        1. `count` - number of exposures, value of metric denominator
        1. `mean` - `sum_value` / `count`
        1. `std` - sample standard deviation
        1. `sum_value` - value of goals, value of metric nominator
        1. `confidence_level` - current confidence level used to calculate `p_value` and `confidence_interval`
        1. `diff` - relative diff between sample means of this and control variant
        1. `test_stat` - value of test statistic of the relative difference in means
        1. `p_value` - p-value of the test statistic under current `confidence_level`
        1. `confidence_interval` - confidence interval of the `diff` under current `confidence_level`
        1. `standard_error` - standard error of the `diff`
        1. `degrees_of_freedom` - degrees of freedom of this variant mean
        """
        stats = stats.transpose(1, 2, 0)

        stat_res = []  # semiresults
        variants_count = stats.shape[0]  # number of variants

        # get only stats (not metric_id, metric_name, exp_variant_id) from the stats array as floats
        stats_values = stats[:, 3:8, :].astype(float)

        # control variant values
        s = stats_values[0]
        count_cont = s[0]  # number of observations
        mean_cont = s[1]  # mean
        std_cont = s[2]  # standard deviation
        conf_level = s[4]  # confidence level

        # this for loop goes over variants and compares one variant values against control variant values for
        # all metrics at once. Similar to scipy.stats.ttest_ind_from_stats
        for i in range(variants_count):
            # treatment variant data
            s = stats_values[i]
            count_treat = s[0]  # number of observations
            mean_treat = s[1]  # mean
            std_treat = s[2]  # standard deviation

            # degrees of freedom
            num = (std_cont ** 2 / count_cont + std_treat ** 2 / count_treat) ** 2
            den = (std_cont ** 4 / (count_cont ** 2 * (count_cont - 1))) + (
                std_treat ** 4 / (count_treat ** 2 * (count_treat - 1))
            )

            with np.errstate(divide="ignore", invalid="ignore"):
                # We fill in zeros, when goal data are missing for some variant.
                # There could be division by zero here which is expected as we return
                # nan or inf values to the caller.
                # np.round() in case of roundoff errors, e.g. f = 9.999999998 => trunc(round(f, 5)) = 10
                f = np.trunc(np.round(num / den, 5))  # (rounded & truncated) degrees of freedom

            # t-quantile
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                t_quantile = st.t.ppf(conf_level + (1 - conf_level) / 2, f)  # right quantile

            # relative difference and test statistics
            with np.errstate(divide="ignore", invalid="ignore"):
                # We fill in zeros, when goal data are missing for some variant.
                # There could be division by zero here which is expected as we return
                # nan or inf values to the caller.
                rel_diff = (mean_treat - mean_cont) / mean_cont
                # standard error for relative difference
                rel_se = (
                    np.sqrt(
                        (mean_treat * std_cont) ** 2 / (mean_cont ** 2 * count_cont) + (std_treat ** 2 / count_treat)
                    )
                    / mean_cont
                )
                test_stat = rel_diff / rel_se

            # p-value
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=RuntimeWarning)
                pval = 2 * (1 - st.t.cdf(np.abs(test_stat), f))

            # confidence interval
            conf_int = rel_se * t_quantile

            # save results
            stat_res.append((rel_diff, test_stat, pval, conf_int, rel_se, f))

        # Tune up results
        s = np.hstack([stats, stat_res])
        s = s.transpose(2, 0, 1)  # move back to metrics, variants, stats order
        x, y, z = s.shape
        arr = s.reshape(x * y, z)

        # Output dataframe
        col = [
            "metric_id",
            "metric_name",
            "exp_variant_id",
            "count",
            "mean",
            "std",
            "sum_value",
            "confidence_level",
            "diff",
            "test_stat",
            "p_value",
            "confidence_interval",
            "standard_error",
            "degrees_of_freedom",
        ]
        r = pd.DataFrame(arr, columns=col)
        return r

    @classmethod
    def multiple_comparisons_correction(
        cls, df: pd.DataFrame, variants: int, metrics: int, confidence_level: float
    ) -> pd.DataFrame:
        """
        [Holm-Bonferroni correction](https://en.wikipedia.org/wiki/Holm%E2%80%93Bonferroni_method)
        for multiple comparisons problem. It is applied when we have more than two variants,
        i.e. we have one control variant and at least two treatment variants.

        It adjusts p-value and length of confidence interval - both to be more conservative.
        [Complete manual](../stats/multiple.md)

        Algorithm:

        For each metric, select (unadjusted) p-values and replace them with adjusted ones.
        Based on adjustment ratio, compute new (adjusted) confidence intervals and replace
        old (unadjusted) ones.

        Arguments:
            df: dataframe as output of [`ttest_evaluation`][epstats.toolkit.statistics.Statistics.ttest_evaluation]
            variants: number of variants in the experiment
            metrics: number of metrics of experiment
            confidence_level: desired confidence level at the end of the experiment, e.g. 0.95

        Returns:
            dataframe of the same format as input with adjusted p-values and confidence intervals.
        """
        alpha = 1 - confidence_level  # level of significance

        for m in range(metrics):
            # indices of rows with metric m data
            index_from = m * variants + 1
            index_to = (m + 1) * variants - 1

            # p-value adjustment
            pvals = df.loc[index_from:index_to, "p_value"].to_list()  # select old p-values
            adj_pvals = multipletests(pvals=pvals, alpha=alpha, method="holm")[1]  # compute adjusted p-values

            # confidence interval adjustment
            # we set ratio to 1 when test_stat is so big that pvals are zero, no reason to update ci
            adj_ratio = np.nan_to_num(pvals / adj_pvals, nan=1)  # adjustment ratio
            adj_alpha = adj_ratio * alpha  # adjusted level alpha

            f = df.loc[index_from:index_to, "degrees_of_freedom"].to_list()  # degrees of freedom
            se = df.loc[index_from:index_to, "standard_error"].to_list()  # standard error

            t_quantile = st.t.ppf(np.ones(variants - 1) - adj_alpha + adj_alpha / 2, f)  # right t-quantile
            adj_conf_int = se * t_quantile  # adjusted confidence interval

            # replace (unadjusted) p-values and confidence intervals with new adjusted ones
            df.loc[index_from:index_to, "p_value"] = adj_pvals
            df.loc[index_from:index_to, "confidence_interval"] = adj_conf_int
        return df

    @classmethod
    def obf_alpha_spending_function(cls, confidence_level: int, total_length: int, actual_day: int) -> int:
        """
        [O'Brien-Fleming alpha spending function](https://online.stat.psu.edu/stat509/node/80/).
        We adjust confidence level in time in experiment. Confidence level in this setting is
        a decreasing function of experiment time. See [Sequential Analysis](../stats/sequential.md) for details.

        Arguments:
            confidence_level: required confidence level at the end of the test, e.g. 0.95
            total_length: length of the test in days, e.g. 7, 14, 21
            actual_day: actual days in the experiment period, must be between 1 and `total_length`

        Returns:
            adjusted confidence level with respect to actual day of the experiment and total
            length of the experiment.
        """
        alpha = 1 - confidence_level
        t = actual_day / total_length  # t in (0, 1]
        q = st.norm.ppf(1 - alpha / 2)  # quantile of normal distribution
        alpha_adj = 2 - 2 * st.norm.cdf(q / np.sqrt(t))
        return 1 - alpha_adj



import pandas as pd

from .experiment import Experiment


class Dao:
    """
    Abstract class interfacing any kind of underlying data source.
    """

    def get_unit_goals(self, experiment: Experiment) -> pd.DataFrame:
        """
        Get goals data pre-aggregated by columns `exp_variant_id`, `unit_type`, `agg_type`, `goal`,
        `dimension`, `dimension_value` and `unit_id`.

        See [`Experiment.evaluate_by_unit`][epstats.toolkit.experiment.Experiment.evaluate_by_unit] for column
        descriptions and example result.
        """
        pass

    def get_agg_goals(self, experiment: Experiment) -> pd.DataFrame:
        """
        Get goals data pre-aggregated by columns `exp_variant_id`, `unit_type`, `agg_type`, `goal`,
        `dimension`, `dimension_value`.

        See [`Experiment.evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_agg] for column
        descriptions and example result.
        """
        pass

    def close(self) -> None:
        """
        Close underlying data source connection and frees resources (if any).
        """
        pass


class DaoFactory:
    """
    Factory creating instances of [`Dao`][epstats.toolkit.dao.Dao] classes.

    It is used in API server to get dao for every request.
    """

    def get_dao(self) -> Dao:
        """
        Create new instance of [`Dao`][epstats.toolkit.dao.Dao] to serve the request.
        """
        pass



from typing import List
import pandas as pd
import numpy as np
from scipy.stats import chisquare

from .parser import Parser


class Check:
    """
    Perform data quality check that accompanies metric evaluation in the experiment.

    See [Data Quality Checks](../stats/basics.md#data-quality-checks) for details about
    data quality checks and [`Evaluation`][epstats.toolkit.experiment.Evaluation] for description of output.
    """

    def __init__(self, id: int, name: str, denominator: str):
        self.id = id
        self.name = name
        self.denominator = denominator
        self._parser = Parser(denominator, denominator)
        self._goals = self._parser.get_goals()

    def get_goals(self) -> List:
        """
        List of all goals needed to evaluate the check in the experiment.

        Returns:
            list of parsed structured goals
        """
        return self._goals

    def evaluate_agg(self, goals: pd.DataFrame, default_exp_variant_id: str) -> pd.DataFrame:
        """
        Evaluate this check from pre-aggregated goals.

        Arguments:
            goals: one row per experiment variant
            default_exp_variant_id: default variant

        See [`Experiment.evaluate_agg`][epstats.toolkit.experiment.Experiment.evaluate_agg] for details
        on `goals` at input.

        Returns:
            `checks` dataframe with columns:

        `checks` dataframe with columns:

        1. `timestamp` - timestamp of evaluation
        1. `exp_id` - experiment id
        1. `check_id` - check id as in [`Experiment`][epstats.toolkit.experiment.Experiment] definition
        1. `variable_id` - name of the variable in check evaluation, SRM check has following variables `p_value`,
        `test_stat`, `confidence_level`
        1. `value` - value of the variable
        """
        raise NotImplementedError()

    def evaluate_by_unit(self, goals: pd.DataFrame, default_exp_variant_id: str) -> pd.DataFrame:
        """
        Evaluate this check from goals aggregated by unit.

        Arguments:
            goals: ne row per experiment variant
            default_exp_variant_id: default variant

        See [`Experiment.evaluate_by_unit`][epstats.toolkit.experiment.Experiment.evaluate_by_unit] for details
        on `goals` at input.

        Returns:
            `checks` dataframe with columns:

        `checks` dataframe with columns:

        1. `timestamp` - timestamp of evaluation
        1. `exp_id` - experiment id
        1. `check_id` - check id as in [`Experiment`][epstats.toolkit.experiment.Experiment] definition
        1. `variable_id` - name of the variable in check evaluation, SRM check has following variables `p_value`,
        `test_stat`, `confidence_level`
        1. `value` - value of the variable
        """
        raise NotImplementedError()


class SrmCheck(Check):
    """
    [Sample ratio mismatch check](../stats/basics.md#sample-ratio-mismatch-check) checking randomization
    of units to variants using [Chi-square test](https://en.wikipedia.org/wiki/Chi-squared_test).
    """

    def __init__(
        self,
        id: int,
        name: str,
        denominator: str,
        confidence_level: float = 0.999,
    ):
        """
        Constructor of the SRM check.

        Arguments:
            id: check (order) id
            name: check name
            denominator: values to check
            confidence_level: confidence level of the statistical test

        Usage:
        ```python
        SrmCheck(1, 'SRM', 'count(test_unit_type.global.exposure)')
        ```
        """
        super().__init__(id, name, denominator)
        self.confidence_level = confidence_level

    def evaluate_agg(self, goals: pd.DataFrame, default_exp_variant_id: str) -> pd.DataFrame:
        """
        See [`Check.evaluate_agg`][epstats.toolkit.check.Check.evaluate_agg].
        """
        # input example:
        # test - srm, a, global.exposure, 10000, 10010, 10010, 0.0, 0.0
        # test - srm, b, global.exposure, 10010, 10010, 10010, 0.0, 0.0
        # test - srm, c, global.exposure, 10040, 10040, 10040, 0.0, 0.0

        # output example:
        # test - srm, 1, SRM, p_value, 0.20438
        # test - srm, 1, SRM, test_stat, 3.17552
        # test - srm, 1, SRM, confidence_level, 0.999

        # prepare data - we only need exposures
        exposures, _, _ = self._parser.evaluate_agg(goals)

        # chi-square test
        with np.errstate(divide="ignore", invalid="ignore"):
            # we fill in zeros, when goal data are missing for some variant.
            # There could be division by zero here which is expected as we return
            # nan or inf values to the caller.
            stat, pval = chisquare(exposures)

        r = pd.DataFrame(
            {
                "check_id": [self.id, self.id, self.id],
                "check_name": [self.name, self.name, self.name],
                "variable_id": ["p_value", "test_stat", "confidence_level"],
                "value": [pval, stat, self.confidence_level],
            }
        )
        return r

    def evaluate_by_unit(self, goals: pd.DataFrame, default_exp_variant_id: str) -> pd.DataFrame:
        """
        See [`Check.evaluate_by_unit`][epstats.toolkit.check.Check.evaluate_by_unit].
        """

        exposures, _, _ = self._parser.evaluate_by_unit(goals)

        # chi-square test
        stat, pval = chisquare(exposures)

        r = pd.DataFrame(
            {
                "check_id": [self.id, self.id, self.id],
                "check_name": [self.name, self.name, self.name],
                "variable_id": ["p_value", "test_stat", "confidence_level"],
                "value": [pval, stat, self.confidence_level],
            }
        )
        return r


class SimpleSrmCheck(SrmCheck):
    """Simplified definition of SRM check."""

    def __init__(
        self,
        id: int,
        name: str,
        denominator: str,
        confidence_level: float = 0.999,
        unit_type: str = "test_unit_type",
    ):
        """
        Constructor of the simplified SRM check.

        It modifies parameter denominator in a way that it is in line with general SRM Check definition. It adds all
        the niceties necessary for proper SrmCheck format. Finaly it calls constructor of the parent SrmCheck class.

        Arguments:
            id: check (order) id
            name: check name
            denominator: value (column) of the denominator
            confidence_level: confidence level of the statistical test
            unit_type: unit type

        Usage:
        ```python
        SimpleSrmCheck(1, 'SRM', 'exposures')
        ```
        """
        agg_type = "global"
        den = "value" + "(" + unit_type + "." + agg_type + "." + denominator + ")"
        super().__init__(id, name, den, confidence_level)



import pandas as pd

import importlib_resources as pkg_resources
from . import resources  # relative-import the *package* containing the templates


class TestData:
    """
    Utility methods to load sample (test) data that are used in unit tests through this
    project.
    """

    @classmethod
    def load_goals_agg(cls, exp_id: str = None) -> pd.DataFrame:
        """
        Load sample of aggregated test data to evaluate metrics. We use this dataset
        in unit testing and we are making it available here for other possible use-cases too.

        See `load_evaluations` set of functions to load corresponding evaluation results.

        Arguments:
            exp_id: experiment id
        """
        df = pd.read_csv(pkg_resources.open_text(resources, "goals_agg.csv"), sep="\t").fillna(
            {"dimension": "", "dimension_value": ""}
        )
        return df[df.exp_id == exp_id] if exp_id is not None else df

    @classmethod
    def load_goals_simple_agg(cls) -> pd.DataFrame:
        """
        Load sample of aggregated test data in simple wide format. File `goals_simple_agg.csv` contains only one
        experiment, so it is sufficient to just open it.

        We use this dataset in unit testing and we are making it available here for other possible use-cases too.

        See `load_evaluations` set of functions to load corresponding evaluation results.
        """
        df = pd.read_csv(pkg_resources.open_text(resources, "goals_simple_agg.csv"), sep="\t")
        return df

    @classmethod
    def load_goals_by_unit(cls, exp_id: str = None) -> pd.DataFrame:
        """
        Load sample of test data by unit to evaluate metrics. We use this dataset
        in unit testing and we are making it available here for other possible use-cases too.

        See `load_evaluations` set of functions to load corresponding evaluation results.

        Arguments:
            exp_id: experiment id
        """
        df = pd.read_csv(pkg_resources.open_text(resources, "goals_by_unit.csv"), sep="\t").fillna(
            {"dimension": "", "dimension_value": ""}
        )
        return df[df.exp_id == exp_id] if exp_id is not None else df

    @classmethod
    def load_evaluations_checks(cls, exp_id: str = None) -> pd.DataFrame:
        """
        Load checks (SRM) evaluations results. This data can be used to do asserts against
        after running evaluation on [pre-aggregated][epstats.toolkit.testing.test_data.TestData.load_goals_agg]
        or [by-unit][epstats.toolkit.testing.test_data.TestData.load_goals_by_unit] test data.

        Arguments:
            exp_id: experiment id
        """
        df = pd.read_csv(pkg_resources.open_text(resources, "evaluations_checks.csv"), sep="\t")
        return df[df.exp_id == exp_id] if exp_id is not None else df

    @classmethod
    def load_evaluations_exposures(cls, exp_id: str = None) -> pd.DataFrame:
        """
        Load exposures evaluations results. This data can be used to do asserts against
        after running evaluation on [pre-aggregated][epstats.toolkit.testing.test_data.TestData.load_goals_agg]
        or [by-unit][epstats.toolkit.testing.test_data.TestData.load_goals_by_unit] test data.

        Arguments:
            exp_id: experiment id
        """
        df = pd.read_csv(pkg_resources.open_text(resources, "evaluations_exposures.csv"), sep="\t")
        return df[df.exp_id == exp_id] if exp_id is not None else df

    @classmethod
    def load_evaluations_metrics(cls, exp_id: str = None) -> pd.DataFrame:
        """
        Load metric evaluations results. This data can be used to do asserts against
        after running evaluation on [pre-aggregated][epstats.toolkit.testing.test_data.TestData.load_goals_agg]
        or [by-unit][epstats.toolkit.testing.test_data.TestData.load_goals_by_unit] test data.

        Arguments:
            exp_id: experiment id
        """
        df = pd.read_csv(pkg_resources.open_text(resources, "evaluations_metrics.csv"), sep="\t")
        return df[df.exp_id == exp_id] if exp_id is not None else df



from pydantic import BaseSettings


class ApiSettings(BaseSettings):
    app_name: str = "ep-stats"
    app_env: str = "dev"
    app_title: str = "Experimentation Platform Statistics"
    app_description: str = "API for statistical evaluation in Experimentation Platform."

    host: str = "0.0.0.0"
    port: int = 8080

    log_level: str = "info"
    evaluation_pool_size: int = 10
    web_workers: int = 1



from .experiment import Experiment, Evaluation, Filter, FilterScope
from .metric import Metric, SimpleMetric
from .check import Check, SrmCheck, SimpleSrmCheck
from .parser import Parser, EpGoal
from .dao import Dao, DaoFactory
from .statistics import Statistics



from setuptools import setup

setup(
    # setup.cfg files are only available in newer version of setuptools
    # this requirement has to be here to cause an error with older versions
    setup_requires=["setuptools >= 40", "wheel >= 0.32", "pytest-runner >= 5.0, <6.0"],
    # Entry point can't be specified in setup.cfg
    entry_points={"console_scripts": ["epstats = epstats.main:main"]},
)



import logging
from typing import Dict
import uvicorn
from statsd import StatsClient
from fastapi import FastAPI, Depends
from fastapi.exception_handlers import (
    http_exception_handler,
    request_validation_exception_handler,
)
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException as StarletteHTTPException

from .api_evaluate import get_evaluate_router
from .json_response import DataScienceJsonResponse
from .api_settings import ApiSettings


def get_api(settings: ApiSettings, get_dao, get_executor_pool, get_statsd) -> FastAPI:
    api = FastAPI(
        title=settings.app_title,
        description=settings.app_description,
        version="0.2.1",
        default_response_class=DataScienceJsonResponse,
    )

    @api.exception_handler(StarletteHTTPException)
    async def custom_http_exception_handler(request, exc):
        """
        We override default exception handler to send exception to the log.
        """
        logger = logging.getLogger("epstats")
        logger.error(f"HttpException status code [{exc.status_code}] detail [{exc.detail}]")
        return await http_exception_handler(request, exc)

    @api.exception_handler(RequestValidationError)
    async def custom_request_validation_exception_handler(request, exc):
        """
        We override default exception handler to send exception to the log.
        """
        logger = logging.getLogger("epstats")
        logger.exception(f"RequestValidationError: [{exc}], [{exc.body}]")
        return await request_validation_exception_handler(request, exc)

    @api.get("/health", tags=["Health"])
    async def readiness_liveness_probe(statsd: StatsClient = Depends(get_statsd)):
        statsd.incr("requests.health")
        return {"message": "ep-stats-api is ready"}

    api.include_router(get_evaluate_router(get_dao, get_executor_pool, get_statsd))

    return api


def serve(api: str, settings: ApiSettings, log_config: Dict):
    logger = logging.getLogger("epstats")
    logger.info(f"Starting {settings.app_name} in env {settings.app_env}")
    logger.info(f"Listening on http://{settings.host}:{settings.port}")
    logger.info(f"Starting with log level {settings.log_level}")
    logger.info(f"Using {settings.web_workers} web server worker threads")
    uvicorn.run(
        api,
        host=settings.host,
        port=settings.port,
        log_level=settings.log_level,
        workers=settings.web_workers,
        timeout_keep_alive=0,
        log_config=log_config,
        http="h11",
    )



import pandas as pd
from .experiment import Experiment


def results_long_to_wide(metrics: pd.DataFrame) -> pd.DataFrame:
    """Adjusts metric resutls from long format to wide."""

    # Compute lower and upper bound for confidence interval
    metrics["conf_int_lower"] = metrics["diff"] - metrics["confidence_interval"]
    metrics["conf_int_upper"] = metrics["diff"] + metrics["confidence_interval"]

    # Change experiment variants to upper case
    metrics = metrics.assign(exp_variant_id=lambda r: r.exp_variant_id.str.upper())

    # Reshape metrics DataFrame - from long to wide
    r = metrics.pivot(
        index=["exp_id", "exp_variant_id"],
        columns=["metric_name", "metric_id"],
        values=["mean", "diff", "conf_int_lower", "conf_int_upper", "p_value"],
    )

    # Add column multiindex names and transpose
    r.columns.names = ["statistic", "metric_name", "metric_id"]
    r = r.transpose()

    # Sort metrics and statistics in the right order
    r.reset_index(inplace=True)
    r["metric_id"] = r.apply(_enrich_metric_id, axis="columns")
    r.sort_values(by="metric_id", inplace=True)
    r.drop(columns=[("metric_id", "")], inplace=True)

    # Set index and transpose back
    r.set_index(["metric_name", "statistic"], inplace=True)
    r = r.transpose()

    return r


def _enrich_metric_id(row):
    """Auxiliary function for proper order of the metrics and statistics."""
    if row[0] == "mean":
        return row[2] + 0.1
    elif row[0] == "diff":
        return row[2] + 0.2
    elif row[0] == "conf_int_lower":
        return row[2] + 0.3
    elif row[0] == "conf_int_upper":
        return row[2] + 0.4
    elif row[0] == "p_value":
        return row[2] + 0.5


def format_results(
    r: pd.DataFrame, experiment: Experiment, format_pct: str = "{:,.1%}", format_pval: str = "{:,.3f}"
) -> pd.DataFrame:
    """
    Method formatting wide dataframe with resutls. Using params `format_pct` and `format_pval` you can set number of
    decimals.

    Arguments:
        r: dataframe with resutls in wide format
        experiment: evaluated experiment
        format_pct: optional param with format of columns `Impact`, `Conf. interval lower bound` and `Conf. interval
        upper bound`
        format_pval: optional param with format of `p-value` column

    Returns:
        nicely formated dataframe
    """
    # Fix ugly naming
    r.rename(
        columns={
            "mean": "Mean",
            "diff": "Impact",
            "conf_int_lower": "Conf. interval lower bound",
            "conf_int_upper": "Conf. interval upper bound",
            "p_value": "p-value",
        },
        inplace=True,
    )

    # Set names for axis
    r.columns.names = ["Metric", "Statistics"]
    r.index.names = ["Experiment Id", "Variant"]

    # How should mean (metric value) formated
    format_mean = [metric.metric_format for metric in experiment.metrics]
    metric_mean_multipliers = [metric.metric_value_multiplier for metric in experiment.metrics]

    # Select appropriate columns
    columns_pct = [col for col in r.columns if ("interval" in col[1]) | ("Impact" in col[1])]
    columns_pvalue = [col for col in r.columns if "p-value" in col[1]]
    columns_mean = [col for col in r.columns if "Mean" in col[1]]

    # Set formatting for specific columns
    columns_pct_format = {col: format_pct for col in columns_pct}
    columns_pvalue_format = {col: format_pval for col in columns_pvalue}
    columns_mean_format = {columns_mean[i]: format_mean[i] for i in range(len(columns_mean))}

    # Apply metric_value_multiplier, e.g. 1000 for RPM
    for i in range(len(columns_mean)):
        r[columns_mean[i]] = r[columns_mean[i]] * metric_mean_multipliers[i]

    # Apply columns formatting including colour p-value format
    return (
        r.style.format(columns_pct_format)
        .format(columns_pvalue_format)
        .format(columns_mean_format)
        .applymap(_p_value_color_format, subset=columns_pvalue)
    )


def _p_value_color_format(pval):
    """Auxiliary function to set p-value color -- green or red."""
    color = "green" if pval < 0.05 else "red"
    return "color: %s" % color
